include "wsdl.grm"

% Define <source> element to wrap around newly constructed operations.
define source
	[SPOFF] '<'source [opt attribute_list] '> [SPON]																	[NL][IN]
		[repeat element] 																								[EX]
	[SPOFF] '</'source> [SPON] 																							[NL]
end define

% Redefine <operation> elements so filename and begin/end line nums are known.
redefine operation 
		[attr srcfilename] [attr srclinenumber] [SPOFF] '< [opt prefix] 'operation [opt attribute_list] [SPON] '> 		[NL][IN]
			[repeat operation_scope] 																					[EX]
		[attr srclinenumber] [SPOFF] '</ [opt prefix] 'operation> [SPON] 												[NL]
	|	[attr srcfilename] [attr srclinenumber] [SPOFF] '< [opt prefix] 'operation [opt attribute_list] '/> [SPON] 		[NL]
end redefine

% Redefine a WSDL program to allow a list of <source> elements.
redefine program 
		...
	|	[repeat source]
end redefine

% Allow all types of elements to be inserted into these during consolidation.
redefine complex_type
		[SPOFF] '< [opt prefix] 'complexType [opt attribute_list] '> [SPON] 											[NL][IN]
			[repeat element] 																							[EX]
		[SPOFF] '</ [opt prefix] 'complexType> [SPON] 																	[NL]
	| 	...
end redefine

redefine simple_type
	[SPOFF] '< [opt prefix] 'simpleType [opt attribute_list] '> [SPON] 													[NL][IN]
		[repeat element] 																								[EX]
	[SPOFF] '</ [opt prefix] 'simpleType> [SPON] 																		[NL]
end redefine			

redefine message
		[SPOFF] '< [opt prefix] 'message [opt attribute_list] [SPON] '> 												[NL][IN]
			[repeat element] 																							[EX]
		[SPOFF] '</ [opt prefix] 'message> [SPON] 																		[NL]
	|	[SPOFF] '< [opt prefix] 'message [opt attribute_list] '/>  [SPON]													[NL]
end redefine

redefine element
		...
	|	[SPOFF] '< [opt prefix] 'element [opt attribute_list] '> [SPON]													[NL][IN]
			[repeat element]																							[EX]
		[SPOFF] '</ [opt prefix] 'element> [SPON]																		[NL]
end redefine

redefine schema
		...
	|	[xml_def]
end redefine
	
% The main function calls the contextualizeOperations function  with each <operation> element,.
% It returns a list of contextualized operations as the new "program."
function main
	replace [program]
		P [program]
		
	% Extract all <operation> elements from the <portTypes> element.
	construct AllPortTypes [repeat port_type]
		_ [^ P]
	construct AllOperations [repeat operation]
		_ [^ AllPortTypes]
		
	% Extract all schema, messages and namespace definitions and export them for use later
	% This saves having to extract the same elements over and over again.
	construct AllSchema [repeat schema]
		_ [^ P]
	construct AllMessages [repeat message]
		_ [^ P]
	deconstruct P
		_ [repeat meta_tag]
		'< _ [opt prefix] 'definitions DefAttrList [opt attribute_list] '>
			_ [repeat top_level_elements] 
		'</ _ [opt prefix] 'definitions>
	construct DefNamespaces [repeat xmlns_attr]
		_ [^ DefAttrList]
	export Schema [repeat schema]
		AllSchema
	export Messages [repeat message]
		AllMessages
	export AllNamespaces [repeat xmlns_attr]
		DefNamespaces
		
	% Construct contextualized operations using the previously extracted <operation> elements
	construct Operations [repeat source]
		_ [contextualizeOperations each AllOperations]
	by
		Operations
end function 

% The contextualizeOperations function takes an <operation> element, contextualizes it, and appends it to a list.
function contextualizeOperations Operation [operation]
	replace [repeat source]
		PreviousOperations [repeat source]
		
	% Deconstruct the operation to get the attribute list filename and begin/end line number
	deconstruct Operation 
		File [srcfilename] BeginLine [srclinenumber] '< _ [opt prefix] 'operation AttrList [opt attribute_list] '> 	
			Contents [repeat operation_scope] 	
		EndLine [srclinenumber] '</ _ [opt prefix] 'operation>
		
	% Add double quotes because these will be used as attributes in the <source> tag
	construct Filename [stringlit]
		_ [quote File]
	construct BeginLineNumber [stringlit]
		_ [quote BeginLine]
	construct EndLineNumber [stringlit]
		_ [quote EndLine]
		
	% Extract the inputs, outputs and faults
	construct Inputs [repeat input]
		_ [^ Contents]
	construct Outputs [repeat output]
		_ [^ Contents]
	construct Faults [repeat fault]
		_ [^ Contents]
	
	% Contextualize the inputs, outputs and faults
	construct NewInputs [repeat operation_scope]
		_ [contextualizeInputs each Inputs]
	construct NewOutputs [repeat operation_scope]
		_ [contextualizeOutputs each Outputs]
	construct NewFaults [repeat operation_scope]
		_ [contextualizeFaults each Faults]
	
	% Append the contextualized  inputs, outputs and faults to use as the contents of the new contextualized operation
	construct NewOperationContents [repeat operation_scope]
		NewInputs [. NewOutputs] [. NewFaults]
		
	% Create the contextualized operation using the newly contextualized inputs, outputs and faults
	construct NewOperation [operation]
		'<'operation AttrList '>
			NewOperationContents
		'</'operation>
		
	% Surround the contextualized operation with <source> tags for NICAD.
	construct SourceTag [source]
		'<'source 'file= Filename 'startline= BeginLineNumber 'endline= EndLineNumber '>
			NewOperation
		'</'source>
	by 
		PreviousOperations [. SourceTag]
end function

% The contextualizeInputs function takes an <input> element, contextualizes it, and appends it to a list.
function contextualizeInputs Input [input]
	replace [repeat operation_scope]
		PreviousInputs [repeat operation_scope]
		
	% Get the messages (already extracted from the main function).
	import Messages [repeat message]
	
	%Get the value of the input's "message" attribute to find the matching one.
	construct AttrList [repeat attribute]
		_ [^ Input]
	construct MessageAttr [repeat message_attr]
		_ [^ AttrList]
	deconstruct MessageAttr
		_ [opt prefix] 'message = MessageName [stringlit] 
		_ [repeat message_attr]
	construct MessageNameNoPrefix [stringlit]
		MessageName [deletePrefix]
	
	% Find the message with the name matching the message attribute and contextualize it.	
	construct ContextualizedMessages [repeat element]
		_ [contextualizeMessagesWithParts MessageNameNoPrefix each Messages]
	construct EmptyMessages [repeat element]
		_ [getEmptyMessages MessageNameNoPrefix each Messages]
	construct NewInputContents [repeat element]
		ContextualizedMessages [. EmptyMessages]
	
	% Create the contextualized input out of the contextualized message.
	construct NewInput [operation_scope]
		'<'input AttrList '>
			NewInputContents
		'</'input>
	by 
		PreviousInputs [. NewInput]
end function 

% The contextualizeOutputs function takes an <output> element, contextualizes it, and appends it to a list.
function contextualizeOutputs Output [output]
	replace [repeat operation_scope]
		PreviousOutputs [repeat operation_scope]
	
	% Get the messages (already extracted from the main function).
	import Messages [repeat message]
	
	%Get the value of the output's "message" attribute to find the matching one.
	construct AttrList [repeat attribute]
		_ [^ Output]
	construct MessageAttr [repeat message_attr]
		_ [^ AttrList]
	deconstruct MessageAttr
		_ [opt prefix] 'message = MessageName [stringlit] 
		_ [repeat message_attr]
	construct MessageNameNoPrefix [stringlit]
		MessageName [deletePrefix]
	
	% Find the message with the name matching the message attribute and contextualize it.	
	construct ContextualizedMessages [repeat element]
		_ [contextualizeMessagesWithParts MessageNameNoPrefix each Messages]
	construct EmptyMessages [repeat element]
		_ [getEmptyMessages MessageNameNoPrefix each Messages]
	construct NewOutputContents [repeat element]
		ContextualizedMessages [. EmptyMessages]
	
	% Create the contextualized output out of the contextualized message.	
	construct NewOutput [operation_scope]
		'<'output AttrList '>
			NewOutputContents
		'</'output>
	by 
		PreviousOutputs [. NewOutput]
end function 

% The contextualizeFaults function takes a <fault> element, contextualizes it, and appends it to a list.
function contextualizeFaults Fault [fault]
	replace [repeat operation_scope]
		PreviousFaults [repeat operation_scope]

	% Get the messages (already extracted from the main function).
	import Messages [repeat message]

	%Get the value of the fault's "message" attribute to find the matching one.
	construct AttrList [repeat attribute]
		_ [^ Fault]
	construct MessageAttr [repeat message_attr]
		_ [^ AttrList]
	deconstruct MessageAttr
		_ [opt prefix] 'message = MessageName [stringlit] 
		_ [repeat message_attr]
	construct MessageNameNoPrefix [stringlit]
		MessageName [deletePrefix]
		
	% Find the message with the name matching the message attribute and contextualize it.	
	construct ContextualizedMessages [repeat element]
		_ [contextualizeMessagesWithParts MessageNameNoPrefix each Messages]
	construct EmptyMessages [repeat element]
		_ [getEmptyMessages MessageNameNoPrefix each Messages]
	construct NewFaultContents [repeat element]
		ContextualizedMessages [. EmptyMessages]

	% Create the contextualized fault out of the contextualized message.
	construct NewFault [operation_scope]
		'<'fault AttrList '>
			NewFaultContents
		'</'fault>
	by 
		PreviousFaults [. NewFault]
end function 

% The contextualizeMessages function takes name and a <message> element and, if the name matches 
% the <message>'s name attribute,  contextualizes it, and appends it to a list.
function contextualizeMessagesWithParts Name [stringlit] Message [message]
	replace [repeat element]
		PreviousMessages [repeat element]
	
	% Get the message's name
	deconstruct Message
		'< _ [opt prefix] 'message AttrList [opt attribute_list] '> 	
			MessageContents [repeat element] 	
		'</ _ [opt prefix] 'message>
	construct NameAttr [repeat name_attr]
		_ [^ AttrList]
	deconstruct NameAttr 
		_ [opt prefix] 'name = MessageName [stringlit] 
		_ [repeat name_attr]
		
	% Match the <message>'s name with the name passed as a parameter.
	% If they do not match, this will fail and the caller will move on to the next <message>.
	deconstruct MessageName
		Name
	
	% Extract and contextualize each <part> element. Parts may refer to <element>'s 
	% or be typed (which may be defined in the <types> element as a <complextType> or <simpleType>).
	% We need to apply different rules for each case.
	construct Parts [repeat part]
		_ [^ MessageContents]
	construct PartsWithElements [repeat element]
		_ [contextualizePartsWithElement each Parts]
	construct PartsWithTypes [repeat element]
		_ [contextualizePartsWithType each Parts]
	construct OtherParts [repeat element]
		_ [getOtherParts each Parts]
	construct NewMessageContents [repeat element]
		PartsWithElements [. PartsWithTypes] [. OtherParts]
		
	% Create the new contextualized <message> with the contextualized <part>. 
	% A special function is used to check if there are any matching elements to add 
	% (sometimes messages contain nothing). If there aren't then no closing tag is used.
	construct Elmt [element]
		'<'message AttrList '/'> 
	construct NewMessage [element]
		Elmt [createMessage NewMessageContents AttrList]
	by 
		PreviousMessages [. NewMessage]
end function




function getEmptyMessages Name [stringlit] Message [message]
	replace [repeat element]
		PreviousMessages [repeat element]
	
	% Get the message's name
	deconstruct Message
		'< _ [opt prefix] 'message AttrList [opt attribute_list] '/'> 	
	construct NameAttr [repeat name_attr]
		_ [^ AttrList]
	deconstruct NameAttr 
		_ [opt prefix] 'name = MessageName [stringlit] 
		_ [repeat name_attr]
		
	% Match the <message>'s name with the name passed as a parameter.
	% If they do not match, this will fail and the caller will move on to the next <message>.
	deconstruct MessageName
		Name
	
		
	% Create the new contextualized <message> with the contextualized <part>. 
	% A special function is used to check if there are any matching elements to add 
	% (sometimes messages contain nothing). If there aren't then no closing tag is used.
	construct NewMessage [element]
		'<'message AttrList '/'> 
	by 
		PreviousMessages [. NewMessage]
end function



% The contextualizePartsWithElement function takes a <part> element as a parameter and 
% contextualizes it by finding the element to which it refers. 
function contextualizePartsWithElement Part [part]
	replace [repeat element]
		PreviousParts [repeat element]
		
	% Get the value of the <part>'s "element" attribute.
	deconstruct Part
		'< _[opt prefix] 'part AttrList [opt attribute_list] '/'>
	construct ElmtAttr [repeat element_attr]
		_ [^ AttrList]
	deconstruct ElmtAttr
		_ [opt prefix] 'element = ElmtName [stringlit] 
		_ [repeat element_attr]
		
	% Get the <element>s from the schema from the namespace matching the prefix of the element name we just found.
	import AllNamespaces [repeat xmlns_attr]
	import Schema [repeat schema]
	construct Prefix [stringlit]
		ElmtName [deleteSuffix]
	construct MatchingNamespace [stringlit]
		_ [getMatchingNamespaceURI Prefix each AllNamespaces]
	construct MatchingSchema [repeat schema]
		_ [getMatchingSchema MatchingNamespace each Schema]
	construct SchemaElements [repeat schema_element]
		_ [^ MatchingSchema]

	% Find the <element> with a name that matches the <part>'s element attribute and contextualize it.
	construct ElementNameNoPrefix [stringlit]
		ElmtName [deletePrefix] 
	construct MatchingElements [repeat element]
		_ [getMatchingSchemaElement ElementNameNoPrefix each SchemaElements]
		
	% Remove some unneccessary <complexType> tags that only clutter the description.
	construct ProcessedMatchingElements [repeat element]
		MatchingElements [stripComplexTypes1] [stripComplexTypes2] [stripComplexTypes3] [stripComplexTypes4] [stripComplexTypes5] [stripComplexTypes6]
		
	% Create the new contextualized <part> with the contextualized <element>. 
	% A special function is used to check if there are any matching elements to add 
	% (schema may be refernced externally, but we don't have access). If there aren't then no closing tag is used.
	construct Elmt [element]
		'<'part AttrList '/'> 
	construct NewPart [element]
		Elmt [createPart ProcessedMatchingElements AttrList]
	by
		PreviousParts [. NewPart]
end function


% The contextualizePartsWithType function takes a <part> element as a parameter and 
% contextualizes it by finding the type to which it refers, if there is one (it may be a primitive type).
function contextualizePartsWithType Part [part]
	replace [repeat element]
		PreviousParts [repeat element]
	
	% Get the value of the <part>'s "type" attribute.	
	deconstruct Part
		'< _[opt prefix] 'part AttrList [opt attribute_list] '/'>
	construct TypeAttr [repeat type_attr]
		_ [^ AttrList]
	deconstruct TypeAttr
		_ [opt prefix] 'type = TypeName [stringlit] 
		_ [repeat type_attr]
		
	% Get the <complexTypes>s from the schema from the namespace matching the prefix of the type name we just found.
	import AllNamespaces [repeat xmlns_attr]
	import Schema [repeat schema]
	construct Prefix [stringlit]
		TypeName [deleteSuffix]
	construct MatchingNamespace [stringlit]
		_ [getMatchingNamespaceURI Prefix each AllNamespaces]
	construct MatchingSchema [repeat schema]
		_ [getMatchingSchema MatchingNamespace each Schema]
	construct ComplexTypes [repeat complex_type]
		_ [^ MatchingSchema]
		
	% Find the <complexType> with a name that matches the <part>'s type attribute and contextualize it.		
	construct TypeNameNoPrefix [stringlit]
		TypeName [deletePrefix] 
	construct MatchingElements [repeat element]
		_ [getMatchingType TypeNameNoPrefix each ComplexTypes]
		
	% Remove some unneccessary <complexType> tags that only clutter the description.
	construct ProcessedMatchingElements [repeat element]
		MatchingElements [stripComplexTypes1] [stripComplexTypes2] [stripComplexTypes3] [stripComplexTypes4] [stripComplexTypes5] [stripComplexTypes6]
		
	% Create the new contextualized <part> with the contextualized <complexType>. 
	% A special function is used to check if there are any matching elements to add 
	% (schema may be refernced externally, but we don't have access). 
	% If there aren't then no closing tag is used.
	construct Elmt [element]
		'<'part AttrList '/'> 	
	construct NewPart [element]
		Elmt [createPart ProcessedMatchingElements AttrList]
	by
		PreviousParts [. NewPart]
end function



function getOtherParts Part [part]
	replace [repeat element]
		PreviousParts [repeat element]
	
	deconstruct Part
		'< _[opt prefix] 'part AttrList [opt attribute_list] '/'>
	construct TypeAttr [repeat type_attr]
		_ [^ AttrList]
	deconstruct not TypeAttr
		_ [opt prefix] 'type = TypeName [stringlit] 
		_ [repeat type_attr]
		
	construct ElmtAttr [repeat element_attr]
		_ [^ AttrList]
	deconstruct not ElmtAttr
		_ [opt prefix] 'element = ElmtName [stringlit] 
		_ [repeat element_attr]
	
	construct NewPart [element]
		'<'part AttrList '/'> 	

	by
		PreviousParts [. NewPart]
end function




% The getMatchingSchemaElement function takes name and a <element> and, if the name matches 
% the <element>'s name attribute,  contextualizes it, and appends it to a list.
function getMatchingSchemaElement Name [stringlit] Elmt [schema_element]
	replace [repeat element]
		PreviousElements [repeat element]
	
	% Get the value of the <element>'s "name" attribute.	
	construct ElmtAttrList [opt attribute_list]
		_ [getSingletonElementAttrList Elmt] [getElementAttrList Elmt]
	construct NameAttr [repeat name_attr]
		_ [^ Elmt]
	construct TypeAttr [repeat type_attr]
		_ [^ Elmt]
	deconstruct NameAttr
		_ [opt prefix] 'name = ElmtName [stringlit] 
		_ [repeat name_attr]
		
	% Match the <element>'s name with the name passed as a parameter.
	% If they do not match, this will fail and the caller will move on to the next <element>.
	deconstruct Name
		ElmtName
	
	% Get all the <complexType> elements.
	import Schema [repeat schema]
	construct SchemaTypes [repeat complex_type]
		_ [^ Schema]

	% We want to extract the <element>s from inside the matching <element> we found so that they can be contextualized. 
	% But the TXL extract function (^) returns the top-level <element> first in the list of extracted <element>s  so we remove it.
	construct ParentAndChildren [repeat schema_element]
		_ [^ Elmt]
	% Remove the top-level <element>.
	deconstruct ParentAndChildren
		_ [schema_element] %  throw away Parent
		Children [repeat schema_element]

	% Contextualize each of the <element>s extracted from the matching <element>
	construct ComplexChildren [repeat element]
		_ [contextualizeTypes each Children]

	% Create a new <element> using the attributes of the original, but containing the newly consolidated children <element>s.
	construct NewElmt [element]
		Elmt
	construct ComplexElement [element]
		NewElmt [createComplexElement ComplexChildren ElmtAttrList] [stripPrefix]
	by
		PreviousElements [. ComplexElement]
end function

% The getMatchingType function is similar to the get MatchingSchemaElement, but for <part>s that refer to a type. 
% It takes name and a <complexType> and, if the name matches the <element>'s type attribute, contextualizes it,
% and appends it to a list.
function getMatchingType Name [stringlit] Elmt [complex_type]
	replace [repeat element]
		PreviousElements [repeat element]
		
	% Get the value of the <complexType>'s "name" attribute.
	deconstruct Elmt 
		'< _ [opt prefix] 'complexType AttrList [opt attribute_list] '> 
			_ [repeat element]
		'</ _ [opt prefix] 'complexType>
	construct NameAttr [repeat name_attr]
		_ [^ Elmt]
	deconstruct NameAttr
		_ [opt prefix] 'name = ElmtName [stringlit] 
		_ [repeat name_attr]
		
	% Match the <complexType>'s name with the name passed as a parameter.
	% If they do not match, this will fail and the caller will move on to the next <complexType>.
	deconstruct Name
		ElmtName
	
	% Extract the <element>s from inside the matching <complexType> we found so that they can be contextualized. 
	construct TypeElements [repeat schema_element]
		_ [^ Elmt]
	
	% Contextualize each of the <element>s extracted from the matching <complexType>.
	construct ContextualizedElements [repeat element]
		_ [contextualizeTypes each TypeElements]
	by
		PreviousElements [. ContextualizedElements]
end function

% The contextualizeTypes function takes a <element> (passed by parameter) and contextualizes it 
% by finding the type to which it refers, if there is one (it may be a primitive type).
% This function is forms a recursive loop with contextualizeComplexTypes function (the function following this one). 
% It calls contextualizeComplexTypes, which calls contextualizeTypes, and so on until no more types can be resolved.
function contextualizeTypes Elmt [schema_element] 
	replace [repeat element]
		PreviousElements [repeat element]
		
	% Get the value of the <element>'s "type" attribute.
	construct ElmtAttrList [opt attribute_list]
		_ [getSingletonElementAttrList Elmt] [getElementAttrList Elmt]
	construct TypeAttr [repeat type_attr]
		_ [^ ElmtAttrList]
	deconstruct TypeAttr
		_ [opt prefix] 'type = TypeString [stringlit] 
		_ [repeat type_attr]
	construct Prefix [stringlit]
		TypeString [deleteSuffix]
	
	% Get the <complexTypes>s and <simpleTypes>s from the schema in the namespace matching the prefix of the type name we just found.
	import AllNamespaces [repeat xmlns_attr]
	import Schema [repeat schema]
	construct MatchingNamespace [stringlit]
		_ [getMatchingNamespaceURI Prefix each AllNamespaces]
	construct MatchingSchema [repeat schema]
		_ [getMatchingSchema MatchingNamespace each Schema]
	construct SchemaComplexTypes [repeat complex_type]
		_ [^ MatchingSchema]
	construct SchemaSimpleTypes [repeat simple_type]
		_ [^ MatchingSchema]
		
	% Contextualize each <element> by finding the matching <complexType> or <simplexType> from the ones just extracted.
	construct ComplexTypes [repeat element]
		_ [contextualizeComplexTypes TypeAttr each SchemaComplexTypes]
	construct SimpleTypes [repeat element]
		_ [contextualizeSimpleTypes TypeAttr each SchemaSimpleTypes]

	% Create a new <element> with the contextualized elements inside.
	construct Types [repeat element]
		SimpleTypes [. ComplexTypes]	
	construct NewElmt [element]
		Elmt
	construct ComplexElement [element]
		NewElmt [createComplexElement Types ElmtAttrList] [stripPrefix]
	by
		PreviousElements [. ComplexElement]
end function

% The contextualizeComplexTypes function takes a "type" attribute (a list of them that should only contain one) 
% and a <complexType> and, if the type matches the <element>'s type attribute, contextualizes it, and appends it to a list.
% This function is forms a recursive loop with contextualizeTypes function (the previous function). 
% It calls contextualizeTypes, which calls contextualizeComplexTypes, and so on until no more types can be resolved.
function contextualizeComplexTypes TypeAttr [repeat type_attr] ComplexType [complex_type]
	replace [repeat element]
		PreviousTypes [repeat element]
	
	% Get the value of the <complexType>'s "name" attribute.
	deconstruct ComplexType
		'< _ [opt prefix] 'complexType AttrList [opt attribute_list] '> 
			Contents [repeat element]
		'</ _ [opt prefix] 'complexType>
	construct TypeNameAttr [repeat name_attr]
		_ [^ AttrList]
	deconstruct TypeNameAttr
		_ [opt prefix] 'name = TypeName [stringlit] 
		_ [repeat name_attr]
	
	% Get the value of the "type" attribute from the one passed, the <complexType> we're trying to find.
	deconstruct TypeAttr
		_ [opt prefix] 'type = ElmtType [stringlit] 
		_ [repeat type_attr]
	construct ElmtTypeNoPrefix [stringlit]
		ElmtType [deletePrefix]

	% Match the <complexType>'s name with the type passed as a parameter.
	% If they do not match, this will fail and the caller will move on to the next <complexType>.
	deconstruct TypeName
		ElmtTypeNoPrefix

	% Extract all of the <element>s from the matching <complexType> and contextualize them.
	construct SchemaElements [repeat schema_element]
		_ [^ Contents]
	construct ComplexElements [repeat element]
		_ [contextualizeTypes each SchemaElements]
	by
		PreviousTypes [. ComplexElements]
end function

% The contextualizeSimpleTypes function takes a "type" attribute (a list of them that should only contain one) 
% and a <simpleType> and, if the type matches the <element>'s type attribute, contextualizes it, and appends it to a list.
function contextualizeSimpleTypes TypeAttr [repeat type_attr] SimpleType [simple_type]
	replace [repeat element]
		PreviousTypes [repeat element]
		
	% Get the value of the <simpleType>'s "name" attribute.
	deconstruct SimpleType
		'< _ [opt prefix] 'simpleType AttrList [opt attribute_list] '> 
			Contents [repeat element]
		'</ _ [opt prefix] 'simpleType>
	construct TypeNameAttr [repeat name_attr]
		_ [^ AttrList]
	deconstruct TypeNameAttr
		_ [opt prefix] 'name = TypeName [stringlit] 
		_ [repeat name_attr]
	
	% Get the value of the "type" attribute from the one passed, the <complexType> we're trying to find.	
	deconstruct TypeAttr
		_ [opt prefix] 'type = ElmtType [stringlit] 
		_ [repeat type_attr]
	construct ElmtTypeNoPrefix [stringlit]
		ElmtType [deletePrefix]

	% Match the <simpleType>'s name with the type passed as a parameter.
	% If they do not match, this will fail and the caller will move on to the next <complexType>.		
	deconstruct TypeName
		ElmtTypeNoPrefix

	% Create a new <element> with the contexts of the matching <simpleType> inside.
	% (Unlike <complexType>s, <simpleType>s do not contain <element>s; they contain some kind of 
	% restriction like ennumeration. Therefore, we don't have to contextualize any further.)
	construct Elmt [element]
		SimpleType
	construct ComplexElement [element]
		Elmt [createComplexElement Contents AttrList] [stripPrefix]
	by
		PreviousTypes [. ComplexElement]
end function

% This helper function replaces a [stringlit] of the form X:Y with Y (deletes the prefix).
% This is used to get the name or type of an element since they will often have prefixes specifying a nampespace.
function deletePrefix 
	replace [stringlit]
		NameWithPrefix [stringlit]
	construct IndexOfColon [number]
		_ [index NameWithPrefix ":"] [+ '1]
	deconstruct not IndexOfColon
		1
	by
		NameWithPrefix [: IndexOfColon 999]
end function

% This helper function replaces a [stringlit] of the form X:Y with X (deletes the suffix).
% Prefixes are used to specify a namespace. This function is used to get the namespace prefix of an element's type
% so that it can find the schema in the same namespace.
function deleteSuffix 
	replace [stringlit]
		NameWithSuffix [stringlit]
	construct IndexOfColon [number]
		_ [index NameWithSuffix ":"] [- '1]
	deconstruct not IndexOfColon
		1
	by
		NameWithSuffix [: 0 IndexOfColon]
end function

% This helper function is to get the attribute list of a singleton element (one that contains no other elements).
% This (along with getElementAttrList) is used sometimes because TXL will extract all attributes, even from elements inside.
% We use this becuase we want to make sure we have the parent element's attributes. And since, we don't know whether the
% element is singleton or not, we apply both; one fails, the other won't.
function getSingletonElementAttrList Elmt [schema_element]
	deconstruct Elmt
		'< _ [opt prefix] 'element AttrList [opt attribute_list] '/> 
	replace [opt attribute_list]
		_ [opt attribute_list]
	by 
		AttrList
end function

% This helper function is to get the attribute list of an element.
% This (along with getSingletonElementAttrList) is used sometimes because TXL will extract all attributes, even from elements inside.
% We use this becuase we want to make sure we have the parent element's attributes. And since, we don't know whether the
% element is singleton or not, we apply both; one fails, the other won't.
function getElementAttrList Elmt [schema_element]
	deconstruct Elmt
		'< _ [opt prefix] 'element AttrList [opt attribute_list] '> 
			_ [repeat schema_scope]
		'</ _ [opt prefix] 'element>
	replace [opt attribute_list]
		_ [opt attribute_list]
	by 
		AttrList
end function

% This helper function takes a list of elements and an attribute list and creates a new <element> 
% with the given attribute list, containing the given elements. 
% We use it so that if there are no elements in the list, we 
function createComplexElement Contents [repeat element] AttrList [opt attribute_list]
	deconstruct Contents
		_ [element]
		_ [repeat element]
	replace [element]
		_ [element]
	by
		'<'element AttrList '>
			Contents
		'</'element>
end function

% This helper function takes a list of elements and an attribute list and creates a new <message> 
% with the given attribute list, containing the given elements. 
function createMessage Contents [repeat element] AttrList [opt attribute_list]
	deconstruct Contents
		_ [element]
		_ [repeat element]
	replace [element]
		_ [element]
	by
		'<'message AttrList '>
			Contents
		'</'message>
end function

% This helper function takes a list of elements and an attribute list and creates a new <part> 
% with the given attribute list, containing the given elements. 
function createPart Contents [repeat element] AttrList [opt attribute_list]
	deconstruct Contents
		_ [element]
		_ [repeat element]
	replace [element]
		_ [element]
	by
		'<'part AttrList '>
			Contents
		'</'part>
end function

% This helper function is applied to <element>'s to remove the namespace prefix ([stringlit] :) from the tag
% so that <element>s from different namespaces can be matched by clone detection. 
% After contextualization, namespaces don't mean anything anymore.
function stripPrefix
	replace [element]
		'< _[opt prefix] 'element AttrList [attribute_list] '/>
	by
		'<'element AttrList '/>
end function

% This helper function matches a namespace prefix to a URI used to find the schema in that namespace.
function getMatchingNamespaceURI Prefix [stringlit] Namespace [xmlns_attr]
	replace [stringlit]
		_ [stringlit]
	deconstruct Namespace 
		'xmlns: PrefixID [id] '= NamespaceURI [stringlit]
	construct PrefixString [stringlit]
		_ [quote PrefixID]
	deconstruct PrefixString 
		Prefix
	by 
		NamespaceURI
end function

% This helper function is used to find the schema from a given namespace, given its URI. 
function getMatchingSchema Namespace [stringlit] Schema [schema]
	replace [repeat schema]
		PreviousSchema [repeat schema]
	construct TargetNamespace [repeat target_namespace_attr]
		_ [^ Schema]
	deconstruct TargetNamespace 
		_ [opt prefix] 'targetNamespace = URI [stringlit] 
		_ [repeat target_namespace_attr]
	deconstruct URI
		Namespace
	by 
		PreviousSchema [. Schema]
end function

% The following helper functions remove <complexType> tags, along with other tags used with them (like <sequence).
% These tags clutter up the operation description and are not necessary for clone detection (and may even result in poor clones). 
rule stripComplexTypes1
	replace [element]
		'< _[opt prefix] 'element AttrList [opt attribute_list] '> 
			'< _[opt prefix] 'complexType>
				'< _[opt prefix] 'sequence>
					Contents [repeat element]
				'</ _[opt prefix] 'sequence>
			'</ _[opt prefix] 'complexType>
		'</ _ [opt prefix] 'element>
	construct NewElmt [element]
		'<'element AttrList '> 
			Contents
		'</'element>
	by
		NewElmt
end rule

rule stripComplexTypes2
	replace [element]
		'< _[opt prefix] 'element AttrList [opt attribute_list] '> 
			'< _[opt prefix] 'complexType>
				'< _[opt prefix] 'sequence>
				'</ _[opt prefix] 'sequence>
			'</ _[opt prefix] 'complexType>
		'</ _ [opt prefix] 'element>
	construct NewElmt [element]
		'<'element AttrList '/> 
	by
		NewElmt
end rule

rule stripComplexTypes3
	replace [element]
		'< _[opt prefix] 'element AttrList [opt attribute_list] '> 
			'< _[opt prefix] 'complexType>
				'< _[opt prefix] 'sequence/>
			'</ _[opt prefix] 'complexType>
		'</ _ [opt prefix] 'element>
	construct NewElmt [element]
		'<'element AttrList '/> 
	by
		NewElmt
end rule

rule stripComplexTypes4
	replace [element]
		'< _[opt prefix] 'element AttrList [opt attribute_list] '> 
			'< _[opt prefix] 'complexType>
				'< _[opt prefix] 'all>
					Contents [repeat element]
				'</ _[opt prefix] 'all>
			'</ _[opt prefix] 'complexType>
		'</ _ [opt prefix] 'element>
	construct NewElmt [element]
		'<'element AttrList '> 
			Contents
		'</'element>
	by
		NewElmt
end rule

rule stripComplexTypes5
	replace [element]
		'< _[opt prefix] 'element AttrList [opt attribute_list] '> 
			'< _[opt prefix] 'complexType>
				'< _[opt prefix] 'all>
				'</ _[opt prefix] 'all>
			'</ _[opt prefix] 'complexType>
		'</ _ [opt prefix] 'element>
	construct NewElmt [element]
		'<'element AttrList '/> 
	by
		NewElmt
end rule

rule stripComplexTypes6
	replace [element]
		'< _[opt prefix] 'element AttrList [opt attribute_list] '> 
			'< _[opt prefix] 'complexType>
				'< _[opt prefix] 'all/>
			'</ _[opt prefix] 'complexType>
		'</ _ [opt prefix] 'element>
	construct NewElmt [element]
		'<'element AttrList '/> 
	by
		NewElmt
end rule
